# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
# Copyright (C) 2025 Northwestern University.
#
# Invenio-App-RDM is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Community sitemap content."""

from invenio_base import invenio_url_for
from invenio_communities.proxies import current_communities
from invenio_search.api import RecordsSearchV2
from invenio_sitemap import SitemapSection, format_to_w3c


class SitemapSectionOfCommunities(SitemapSection):
    """Defines the Sitemap entries for Communities."""

    def iter_entities(self):
        """Iterate over objects."""
        communities_scan = (
            RecordsSearchV2(index=current_communities.service.record_cls.index._name)
            .filter("term", **{"access.visibility": "public"})
            .filter("term", deletion_status="P")
            .sort("-updated")
            # Using preserve_order is fine.
            # Using Point in Time is a recommended alternative but not
            # particularly better than this one for our needs. See
            # https://opensearch.org/docs/latest/search-plugins/searching-data/point-in-time/
            .params(preserve_order=True)
            # In keeping with original: only page is looked for
            # (curation policy is not most relevant)
            .source(["slug", "updated", "metadata.page"])
            .scan()
        )

        for community in communities_scan:
            yield {
                "slug": community.slug,
                "updated": community.updated,
                "loc": invenio_url_for(
                    "invenio_app_rdm_communities.communities_detail",
                    pid_value=community.slug,
                ),
            }

            if "page" in community.get("metadata", {}):
                yield {
                    "slug": community.slug,
                    "updated": community.updated,
                    "loc": invenio_url_for(
                        "invenio_communities.communities_about",
                        pid_value=community.slug,
                    ),
                }

    def to_dict(self, entity):
        """To dict used in sitemap."""
        return {
            "loc": entity["loc"],
            "lastmod": format_to_w3c(entity["updated"]),
        }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 CERN.
#
# Invenio-App-RDM is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Record migration script from InvenioRDM 13.0 to 14.0.

Disclaimer: This script is intended to be executed *only once*, namely when
upgrading from InvenioRDM 13.0 to 14.0!
If this script is executed at any other time, probably the best case scenario
is that nothing happens!


This script has been tested for the following scenarios:
1. Draft with resource type publication-thesis
2. Record with resource type publication-thesis with a DOI and no draft
3. Record with resource type publication-thesis with a no DOI and an existing draft
4. Records with multiple versions
"""

import time

from click import secho
from invenio_access.permissions import system_identity
from invenio_db import db
from invenio_drafts_resources.resources.records.errors import DraftNotCreatedError
from invenio_rdm_records.proxies import current_rdm_records_service as records_service
from invenio_search.engine import dsl


def run_upgrade(has, migrate_record, migrate_draft):
    """Run upgrade on selected records and drafts.

    Args:
        has (dsl.Q): Query filter to select records/drafts to update.
        migrate_record (callable): Function to migrate a record.
        migrate_draft (callable): Function to migrate a draft.
    """
    # Handle published records
    published_records = records_service.scan(
        identity=system_identity,
        params={"allversions": True},
        extra_filter=has,
    )
    for result in published_records.hits:
        try:
            migrate_record(result)
        except Exception as error:
            secho(f"> Error {repr(error)}", fg="red")
            error = f"Record {result['id']} failed to update"

    # Handle draft records
    draft_records = records_service._search(
        identity=system_identity,
        action="scan",
        params={"allversions": True},
        search_preference=None,
        record_cls=records_service.draft_cls,
        search_opts=records_service.config.search_drafts,
        extra_filter=has,
        permission_action="read_draft",
    ).scan()
    for result in draft_records:
        try:
            migrate_draft(result)
        except Exception as error:
            secho(f"> Error {repr(error)}", fg="red")
            error = f"Draft {result['id']} failed to update"


def run_update_for_resource_type():
    """Run update for resource type."""

    def migrate_resource_type_in_record(hit_result):
        """
        Update resource type from publication-thesis to publication-dissertation.

        We go through the service layer to automatically trigger the DOI update and re-indexing.
        """
        secho(f"Updating resource type for record {hit_result['id']}", fg="yellow")
        record = records_service.read(system_identity, hit_result["id"])
        if record.data["metadata"]["resource_type"]["id"] != "publication-thesis":
            secho(
                f"Skipping record <{record.id}> because it doesn't have resource-type 'publication-thesis'!",
                fg="yellow",
            )
            return

        try:
            draft = records_service.read_draft(system_identity, record.id)
            # Step 1: Update the resource type in the record via low-level API
            # We need to make sure we don't publish the record with different metadata
            secho(
                f"Record <{record.id}> has an existing draft <{draft.id}>! Updating record via low-level API.",
                fg="yellow",
            )
            # Update the record directly without affecting the draft
            record._record["metadata"]["resource_type"][
                "id"
            ] = "publication-dissertation"
            # Save the record changes and reindex
            record._record.commit()
            db.session.commit()
            records_service.indexer.index(record._record)
            # Update DOI metadata if record has DOI
            if hasattr(record, "pids") and record.pids.get("doi", None):
                records_service.pids.register_or_update(
                    system_identity, record.id, "doi", parent=False
                )
            # Step 2: Update the resource type in the draft
            secho(f"Updating resource type for draft {draft.id}", fg="yellow")
            draft.data["metadata"]["resource_type"]["id"] = "publication-dissertation"
            # After updating the record, update the draft's fork_version_id to match the record's new version_id, to avoid conflicts when publishing
            draft._record.fork_version_id = record._record.revision_id
            updated_draft = records_service.update_draft(
                system_identity, draft.id, draft.data
            )
            secho(f"Draft {draft.id} has been updated successfully.", fg="green")
        except DraftNotCreatedError:
            # If the draft didn't exist, we simply edit and publish the record
            draft = records_service.edit(system_identity, record.id)
            draft.data["metadata"]["resource_type"]["id"] = "publication-dissertation"
            updated_draft = records_service.update_draft(
                system_identity, draft.id, draft.data
            )
            record = records_service.publish(system_identity, updated_draft.id)

        secho(f"Record <{record.id}> has been updated successfully.", fg="green")

    def migrate_resource_type_in_draft(hit_result):
        """
        Update resource type from publication-thesis to publication-dissertation.

        We go through the service layer to automatically trigger the DOI update and re-indexing.
        """
        secho(f"Updating resource type for draft {hit_result['id']}", fg="yellow")
        draft = records_service.edit(system_identity, hit_result["id"])
        if draft.data["metadata"]["resource_type"]["id"] != "publication-thesis":
            secho(
                f"Skipping draft <{draft.id}> because it doesn't have resource-type 'publication-thesis'!",
                fg="yellow",
            )
            return

        draft.data["metadata"]["resource_type"]["id"] = "publication-dissertation"
        updated_draft = records_service.update_draft(
            system_identity, draft.id, draft.data
        )
        secho(f"Draft <{updated_draft.id}> has been updated successfully.", fg="green")

    # Query records/drafts with resource type publication-thesis
    has_resource_type = dsl.Q(
        "query_string", query="metadata.resource_type.id:publication-thesis"
    )

    secho("Resource type update has started.", fg="green")

    run_upgrade(
        has_resource_type,
        migrate_resource_type_in_record,
        migrate_resource_type_in_draft,
    )

    secho("Resource type update has finished.", fg="green")


def execute_upgrade():
    """Execute the upgrade from InvenioRDM 13.0 to 14.0.

    Please read the disclaimer on this module before thinking about executing
    this function!
    THIS MODULE IS WORK IN PROGRESS, UNTIL official v14 release

    NOTE:
    since the data upgrade steps are more selective now, the approach how to do
    it has been changed. now the records/drafts which should be updated are
    searched by a filter and then the updates are applied to those
    records/drafts explicitly. this should improve speed and should make it
    easier to upgrade large instances

    """
    secho("Starting data migration...", fg="green")

    run_update_for_resource_type()


# if the script is executed on its own, perform the upgrade
if __name__ == "__main__":
    execute_upgrade()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2025 CERN.
# Copyright (C) 2024-2025 Graz University of Technology.
# Copyright (C) 2025 Northwestern University.
#
# Invenio-App-RDM is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Record migration script from InvenioRDM 12.0 to 13.0.

Disclaimer: This script is intended to be executed *only once*, namely when
upgrading from InvenioRDM 12.0 to 13.0!
If this script is executed at any other time, probably the best case scenario
is that nothing happens!


This script has been tested with following data:

- record
    - internal_notes
"""

import sys
import traceback

from click import secho
from invenio_access.permissions import system_identity
from invenio_db import db
from invenio_rdm_records.proxies import current_rdm_records_service as records_service
from invenio_search import current_search_client as search_client
from invenio_search.engine import dsl
from invenio_search.utils import prefix_index
from invenio_vocabularies.contrib.affiliations.api import Affiliation
from invenio_vocabularies.contrib.names.api import Name
from sqlalchemy import select


def run_upgrade(has, migrate):
    """Run upgrade."""
    record_success_counter = 0
    record_error_counter = 0
    draft_success_counter = 0
    draft_error_counter = 0

    # Handle published records
    published_records = records_service.search(
        system_identity,
        params={"allversions": True, "include_deleted": True},
        extra_filter=has,
    )
    for result in published_records.hits:
        record = records_service.record_cls.pid.resolve(result["id"])
        try:
            migrate(record)
            record_success_counter += 1
        except Exception as error:
            secho(f"> Error {repr(error)}", fg="red")
            error = f"Record {record.pid.pid_value} failed to update"
            record_error_counter += 1

    # Handle draft records
    draft_records = records_service.search_drafts(
        system_identity,
        params={"allversions": True},
        extra_filter=has,
    )
    for result in draft_records.hits:
        draft = records_service.draft_cls.pid.resolve(
            result["id"],
            registered_only=False,
        )
        try:
            migrate(draft)
            draft_success_counter += 1
        except Exception as error:
            secho(f"> Error {repr(error)}", fg="red")
            error = f"Draft {draft.pid.pid_value} failed to update"
            draft_error_counter += 1

    if draft_error_counter > 0 or record_error_counter > 0:
        db.session.rollback()
        secho(
            f"Migration failed: {record_error_counter} records had failures and {draft_error_counter} drafts had failures",
            fg="red",
        )
        secho(
            "The changes have been rolled back. Please fix the above listed errors and try the upgrade again",
            fg="yellow",
            err=True,
        )
    elif draft_success_counter > 0 or record_success_counter > 0:
        db.session.commit()
        secho(
            f"Migration completed: {record_success_counter} records have been updated and {draft_success_counter} drafts have been updated",
            fg="green",
        )
    else:
        secho(
            "Migration completed: no records or drafts required updating.", fg="green"
        )


def run_upgrade_for_thesis():
    """Run upgrade for thesis."""

    def migrate_thesis_university(record_or_draft):
        custom_fields = record_or_draft.get("custom_fields", {})
        university = custom_fields.get("thesis:university")
        if university and "thesis:thesis" not in custom_fields:
            custom_fields["thesis:thesis"] = {"university": university}

        record_or_draft.commit()

    # Common query filter
    has_thesis = dsl.Q("exists", field="custom_fields.thesis:university")

    secho("Thesis upgrade has started.", fg="green")

    run_upgrade(has_thesis, migrate_thesis_university)

    secho("Thesis upgrade has finished.", fg="green")


def run_upgrade_for_affiliations():
    """Update affiliations entry so that they conform to new shape."""
    secho("Affiliations upgrade has started.", fg="green")

    error = False
    # Batch intake to limit memory usage
    stmt = select(Affiliation.model_cls).execution_options(yield_per=250)

    for affiliation_model in db.session.scalars(stmt):
        try:
            data_for_affiliation = affiliation_model.data
            data_for_affiliation.pop("id", None)
            data_for_affiliation.pop("pid", None)
            affiliation = Affiliation(data_for_affiliation, model=affiliation_model)
            affiliation.commit()
        except Exception as e:
            secho(f"Migration failed with '{repr(e)}'.", fg="red")
            secho(f"Affiliation {affiliation_model.pid} failed to update", fg="red")
            trace = traceback.format_exc()
            secho(f"Traceback {trace}", fg="red")
            error = True
            break

    if error:
        db.session.rollback()
        secho("Affiliations upgrade failed.", fg="red")
        sys.exit(1)
    else:
        db.session.commit()
        secho("Affiliations upgrade succeeded.", fg="green")


def run_upgrade_for_names():
    """Update names entry so that they conform to new shape."""
    secho("Names upgrade has started.", fg="green")

    error = False
    # Batch intake to limit memory usage
    stmt = select(Name.model_cls).execution_options(yield_per=250)

    for name_model in db.session.scalars(stmt):
        try:
            data_for_name = name_model.data
            data_for_name.pop("id", None)
            data_for_name.pop("pid", None)
            name = Name(data_for_name, model=name_model)
            name.commit()
        except Exception as e:
            secho(f"Migration failed with '{repr(e)}'.", fg="red")
            secho(f"Name {name_model.pid} failed to update", fg="red")
            trace = traceback.format_exc()
            secho(f"Traceback {trace}", fg="red")
            error = True
            break

    if error:
        db.session.rollback()
        secho("Names upgrade failed.", fg="red")
        sys.exit(1)
    else:
        db.session.commit()
        secho("Names upgrade succeeded.", fg="green")


def run_upgrade_for_event_stats_mappings():
    """Update the live event stats mappings to add missing fields."""
    secho("Event stats mappings upgrade has started.", fg="green")

    # Find the latest mappings for views and download stats events
    for event_type in ("record-view", "file-download"):
        try:
            events_index = prefix_index(f"events-stats-{event_type}-*")
            res = search_client.indices.get(events_index)
            last_two_indices = sorted(res.keys())[-2:]
            for index in last_two_indices:
                res = search_client.indices.put_mapping(
                    index=index,
                    body={"properties": {"is_machine": {"type": "boolean"}}},
                )
        except Exception as e:
            secho(f"Mapping update for {event_type} failed with '{repr(e)}'.", fg="red")
            trace = traceback.format_exc()
            secho(f"Traceback {trace}", fg="red")
    secho("Event stats mappings upgrade succeeded.", fg="green")


def execute_upgrade():
    """Execute the upgrade from InvenioRDM 12.0 to 13.0.0.

    Please read the disclaimer on this module before thinking about executing
    this function!

    NOTE:
    since the data upgrade steps are more selective now, the approach how to do
    it has been changed. now the records/drafts which should be updated are
    searched by a filter and then the updates are applied to those
    records/drafts explicitly. this should improve speed and should make it
    easier to upgrade large instances

    """
    secho("Starting data migration...", fg="green")

    run_upgrade_for_thesis()
    run_upgrade_for_affiliations()
    run_upgrade_for_names()
    run_upgrade_for_event_stats_mappings()


# if the script is executed on its own, perform the upgrade
if __name__ == "__main__":
    execute_upgrade()
